unit DiggerFont;

{$ifdef USE_RANGE_CHECKING}
  {$RANGECHECKS ON}
{$else}
  {$RANGECHECKS OFF}
{$endif}

interface

uses
  Borland.Delphi.Types,
  Oblong.Types;

const
  NumFontChars = 59;

type
  TFont = class
  private
    Position: TPoint;
    Size: TSize;
    Characters: array[0..NumFontChars - 1] of TBitmap;
  public
    constructor Create(Bitmap: TBitmap; ASize: TSize);
    procedure Reset;
    procedure Write(Graphics: TGraphics; const Str: String);
    procedure NewLine;
  end;

implementation

{ TFont }

constructor TFont.Create(Bitmap: TBitmap; ASize: TSize);
var
  I: Integer;
  Cursor: TRect;
begin
  inherited Create;
  Reset;
  Size := ASize;
  Cursor := TRect.Create(0, 0, Size.Width, Size.Height);
  for I := 0 to NumFontChars - 1 do
  begin
    Characters[I] := Bitmap.Clone(Cursor, TPixelFormat.Format24bppRgb);
    Characters[I].MakeTransparent(TColor.FromArgb(0, 0, 0));
    Cursor.Y := Cursor.Y + Cursor.Height
  end
end;

procedure TFont.Reset;
begin
  Position := TPoint.Create(0, 0)
end;

procedure TFont.Write(Graphics: TGraphics; const Str: String);
var
  I: Integer;
begin
  //for I := 0 to Length(Str) - 1 do
  for I := 1 to Length(Str) do
  begin
    Graphics.DrawImageUnscaled(
      Characters[Byte(Ord(Str[I]) - 32)], Position.X, Position.Y);
    Position.X := Position.X + Size.Width
  end
end;

procedure TFont.NewLine;
begin
  Position.X := 0;
  Position.Y := Position.Y + Size.Height
end;

end.
