unit DiggerLevel;

{$ifdef USE_RANGE_CHECKING}
  {$RANGECHECKS ON}
{$else}
  {$RANGECHECKS OFF}
{$endif}

{$ifdef LINK_RESOURCES}
  //Change this directive once managed resource linking is properly supported
  //{$R Level.bin}
{$endif}

{
Level data:
  there are 30 levels
  each level takes 156 bytes
  the level data is 30 * 156 = 4680 bytes
  Data:
    bytes 0 to 139 ($8B)- cell data
      one cell in the high nybble
      next cell in the low nybble
    byte 145 ($91) - X coord of digger start position
    byte 146 ($92) - Y coord of digger start position + 2
    byte 147 ($93) - number of diamonds required before exit is allowed
      (e.g. level 1 is $53 -> 53 diamonds required)
    bytes 148 ($94) to 155 ($9B) - ghost data
      one ghost data in high nybble
      next ghost data in low nybble
}

interface

uses
  Oblong.Types,
  Oblong.VCL.Controls;

const
  LeftKey = 0;
  RightKey = 1;
  UpKey = 2;
  DownKey = 3;
  HighestLevel = 30;
  NumDirKeys = 4;
  StartTime = 5000;
  DiamondPoints = 3;
  GhostKillPoints = 99;
  NumGhosts = 16;
  LevelDataSize = 156;
  MapSizeX = 20;
  MapSizeY = 14;

  SpriteNull = 0;
  SpriteStone = 1;
  SpriteGround = 2;
  SpriteGhostDown = 3;
  SpriteGhostLeft = 4;
  SpriteGhostRight = 5;
  SpriteGhostUp = 6;
  SpriteDiamond = 7;
  SpriteDiamondBlink1 = 8;
  SpriteDiamondBlink2 = 9;
  SpriteDiamondBlink3 = 10;
  SpriteDiamondBlink4 = 11;
  SpriteDiamondBlink5 = 12;
  SpriteDiamondBlink6 = 13;
  SpriteWall = 14;
  SpriteDiggerCenter1 = 15;
  SpriteDiggerLeft1 = 16;
  SpriteDiggerLeft2 = 17;
  SpriteDiggerLeft3 = 18;
  SpriteDiggerLeft4 = 19;
  SpriteDiggerRight1 = 20;
  SpriteDiggerRight2 = 21;
  SpriteDiggerRight3 = 22;
  SpriteDiggerRight4 = 23;
  SpriteDiggerUp1 = 24;
  SpriteDiggerUp2 = 25;
  SpriteDiggerDown1 = 26;
  SpriteDiggerDown2 = 27;
  SpriteDiggerCenter2 = 28;
  SpriteDigger15 = 29;
  SpriteDigger16 = 30;
  SpriteDiggerDead = 31;
  SpriteExit = 32;
  SpriteChanger = 33;

  NumDiamondBlinks = 6;
  NumDiggerLeftBmps = 6;
  NumDiggerRightBmps = 6;
  NumDiggerDownBmps = 2;
  NumDiggerUpBmps = 2;
  NumDiggerCenterBmps = 30;
  LastDiamondBlink = SpriteDiamondBlink6;

type
  TDirections = (None, Left, Right, Up, Down);

  TPosition = class
  public
    X, Y: Integer;
    constructor Create; overload;
    constructor Create(X, Y: Integer); overload;
    constructor Create(Pos: TPosition); overload;
    function Equal(Pos: TPosition): Boolean;
  end;

  TSprite = class
  public
    Invalidate: Boolean;
    constructor Create;
    function GetBitmap: Integer; virtual;
  end;

  TGround = class(TSprite)
  public
    function GetBitmap: Integer; override;
  end;

  TLevel = class;

  TGhost = class(TSprite)
  public
    Direction: TDirections; //defaults to None
    Dead: Boolean; //defaults to False
    function GetBitmap: Integer; override;
    procedure Move(L: TLevel); virtual; abstract;
    procedure Die;
    procedure Blast(L: TLevel);
  end;

  TGhost180 = class(TGhost)
  public
    procedure Move(L: TLevel); override;
  end;

  TGhost90L = class(TGhost)
  public
    procedure Move(L: TLevel); override;
  end;

  TGhost90LR = class(TGhost)
  public
    LastTurn: TDirections; //defaults to None
    procedure Move(L: TLevel); override;
  end;

  TGhost90R = class(TGhost)
  public
    procedure Move(L: TLevel); override;
  end;

  TWall = class(TSprite)
  public
    function GetBitmap: Integer; override;
  end;

  TDigger = class(TSprite)
  public
    Keys: array[0..NumDirKeys-1] of Boolean; //which keys are currently down
//    Stone: array[0..1] of Boolean;
    Direction: TDirections; //defaults to None
    Step: Integer; //defaults to 0
    Dead: Boolean; //defaults to False;
    constructor Create;
    function GetBitmap: Integer; override;
    procedure Animate(L: TLevel; X, Y: Integer);
    procedure Move(L: TLevel);
    procedure Die;
  end;

  TExit = class(TSprite)
  public
    function GetBitmap: Integer; override;
  end;

  TChanger = class(TSprite)
  public
    function GetBitmap: Integer; override;
  end;

  TGravitation = class(TSprite)
  public
    procedure Simulate(L: TLevel; X, Y: Integer);
    procedure Move(L: TLevel; X, Y: Integer);
  end;

  TMarker = class(TGravitation);

  TStone = class(TGravitation)
  public
    function GetBitmap: Integer; override;
  end;

  TUvStone = class(TGravitation);

  TDiamond = class(TGravitation)
  public
    Counter: Integer; //defaults to 0
    procedure Blink;
    function GetBitmap: Integer; override;
  end;

  TNothing = class(TSprite);

  TBuffer = class(TSprite);

  TLevel = class
  private
    Data: array of Byte;
  public
    Digger: TDigger;
    Number,
    Diamonds,
    Collected,
    Time,
    Score: Integer; //these all default to 0
    Map: array[0..MapSizeX-1, 0..MapSizeY-1] of TSprite;
    Ghosts: array[0..NumGhosts-1] of TGhost;
    constructor Create;
    procedure Load;
    procedure ExitLevel;
    procedure CreateSprite(X, Y, ID: Integer);
    procedure SetSprite(X, Y: Integer; S: TSprite);
    function GetSpritePosition(S: TSprite): TPosition;
    procedure AnimateDiamonds;
    procedure Tick;
  end;

implementation

{$ifdef USE_CLASS_REFS}
uses
  Oblong.Vcl.Dialogs;
{$endif}

{ TPosition }

constructor TPosition.Create;
begin
  inherited;
  X := 0;
  Y := 0
end;

constructor TPosition.Create(X, Y: Integer);
begin
  inherited Create;
  Self.X := X;
  Self.Y := Y
end;

constructor TPosition.Create(Pos: TPosition);
begin
  inherited Create;
  Self.X := Pos.X;
  Self.Y := Pos.Y
end;

function TPosition.Equal(Pos: TPosition): Boolean;
begin
  inherited;
  Result := (Pos.X = Self.X) and (Pos.Y = Self.Y)
end;

{ TSprite }

constructor TSprite.Create;
begin
  inherited;
  Invalidate := True
end;

function TSprite.GetBitmap: Integer;
begin
  Result := SpriteNull
end;

{ TGround }

function TGround.GetBitmap: Integer;
begin
  Result := SpriteGround
end;

{ TGhost }

procedure TGhost.Blast(L: TLevel);
var
  CurrentPos: TPosition;
  X, Y: Integer;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  for Y := CurrentPos.Y - 1 to CurrentPos.Y + 1 do
    for X := CurrentPos.X - 1 to CurrentPos.X + 1 do
    begin
      if (X in [0..MapSizeX-1]) and (Y in [0..MapSizeY-1]) then
      begin
        if L.Map[X, Y] is TDigger then
          L.Digger.Die
        else
        begin
          if L.Map[X, Y] is TGhost then
          begin
            TGhost(L.Map[X, Y]).Die;
            Inc(L.Score, GhostKillPoints)
          end;
          L.SetSprite(X, Y, TNothing.Create)
        end
      end
    end;
  Die
end;

procedure TGhost.Die;
begin
  Dead := True
end;

function TGhost.GetBitmap: Integer;
begin
  case Direction of
    Left: Result := SpriteGhostLeft;
    Right: Result := SpriteGhostRight;
    Up: Result := SpriteGhostUp;
  else
    Result := SpriteGhostDown;
  end
end;

{ TGhost180 }

procedure TGhost180.Move(L: TLevel);
var
  CurrentPos, NewPos: TPosition;
  NewPosTry: array[0..3] of TPosition;
  I: Integer;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  for I := Low(NewPosTry) to High(NewPosTry) do
    NewPosTry[I] := TPosition.Create(CurrentPos);
  if Direction = Left then
  begin
    Dec(NewPosTry[0].X);
    Inc(NewPosTry[1].X)
  end
  else if Direction = Right then
  begin
    Inc(NewPosTry[0].X);
    Dec(NewPosTry[1].X)
  end
  else if Direction = Up then
  begin
    Dec(NewPosTry[0].Y);
    Inc(NewPosTry[1].Y)
  end
  else if Direction = Down then
  begin
    Inc(NewPosTry[0].Y);
    Dec(NewPosTry[1].Y)
  end;
  for I := Low(NewPosTry) to High(NewPosTry) do
    if not CurrentPos.Equal(NewPosTry[I]) then
    begin
      NewPos := TPosition.Create(NewPosTry[I]);
      // Digger
      if L.Map[NewPos.X, NewPos.Y] is TDigger then
        L.Digger.Die;
      // Nothing
      if L.Map[NewPos.X, NewPos.Y] is TNothing then
      begin
        if NewPos.X < CurrentPos.X then
          Direction := Left
        else if NewPos.X > CurrentPos.X then
          Direction := Right;
        if NewPos.Y < CurrentPos.Y then
          Direction := Up
        else if NewPos.Y > CurrentPos.Y then
          Direction := Down;
        L.SetSprite(NewPos.X, NewPos.Y, Self);
        L.SetSprite(CurrentPos.X, CurrentPos.Y, TNothing.Create);
        Break
      end
    end
end;

{ TGhost90L }

procedure TGhost90L.Move(L: TLevel);
var
  CurrentPos, NewPos: TPosition;
  NewPosTry: array[0..3] of TPosition;
  I: Integer;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  for I := Low(NewPosTry) to High(NewPosTry) do
    NewPosTry[I] := TPosition.Create(CurrentPos);
  if Direction = Left then
  begin
    Dec(NewPosTry[0].X);
    Inc(NewPosTry[1].Y);
    Dec(NewPosTry[2].Y);
    Inc(NewPosTry[3].X);
  end
  else if Direction = Right then
  begin
    Inc(NewPosTry[0].X);
    Dec(NewPosTry[1].Y);
    Inc(NewPosTry[2].Y);
    Dec(NewPosTry[3].X);
  end
  else if Direction = Up then
  begin
    Dec(NewPosTry[0].Y);
    Dec(NewPosTry[1].X);
    Inc(NewPosTry[2].X);
    Inc(NewPosTry[3].Y);
  end
  else if Direction = Down then
  begin
    Inc(NewPosTry[0].Y);
    Inc(NewPosTry[1].X);
    Dec(NewPosTry[2].X);
    Dec(NewPosTry[3].Y);
  end;
  for I := Low(NewPosTry) to High(NewPosTry) do
    if not CurrentPos.Equal(NewPosTry[I]) then
    begin
      NewPos := TPosition.Create(NewPosTry[I]);
      // Digger
      if L.Map[NewPos.X, NewPos.Y] is TDigger then
        L.Digger.Die;
      // Nothing
      if L.Map[NewPos.X, NewPos.Y] is TNothing then
      begin
        if NewPos.X < CurrentPos.X then
          Direction := Left
        else if NewPos.X > CurrentPos.X then
          Direction := Right;
        if NewPos.Y < CurrentPos.Y then
          Direction := Up
        else if NewPos.Y > CurrentPos.Y then
          Direction := Down;
        L.SetSprite(NewPos.X, NewPos.Y, Self);
        L.SetSprite(CurrentPos.X, CurrentPos.Y, TNothing.Create);
        Break
      end
    end
end;

{ TGhost90LR }

procedure TGhost90LR.Move(L: TLevel);
var
  CurrentPos, NewPos: TPosition;
  NewPosTry: array[0..3] of TPosition;
  I: Integer;
  LastDirection: TDirections;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  for I := Low(NewPosTry) to High(NewPosTry) do
    NewPosTry[I] := TPosition.Create(CurrentPos);
  if Direction = Left then
  begin
    Dec(NewPosTry[0].X);
    Inc(NewPosTry[3].X);
    if LastTurn = Left then
    begin
      Dec(NewPosTry[1].Y);
      Inc(NewPosTry[2].Y)
    end
    else
    begin
      Inc(NewPosTry[1].Y);
      Dec(NewPosTry[2].Y)
    end
  end
  else if Direction = Right then
  begin
    Inc(NewPosTry[0].X);
    Dec(NewPosTry[3].X);
    if LastTurn = Left then
    begin
      Inc(NewPosTry[1].Y);
      Dec(NewPosTry[2].Y)
    end
    else
    begin
      Dec(NewPosTry[1].Y);
      Inc(NewPosTry[2].Y)
    end
  end
  else if Direction = Up then
  begin
    Dec(NewPosTry[0].Y);
    Inc(NewPosTry[3].Y);
    if LastTurn = Left then
    begin
      Inc(NewPosTry[1].X);
      Dec(NewPosTry[2].X)
    end
    else
    begin
      Dec(NewPosTry[1].X);
      Inc(NewPosTry[2].X)
    end
  end
  else if Direction = Down then
  begin
    Inc(NewPosTry[0].Y);
    Dec(NewPosTry[3].Y);
    if LastTurn = Left then
    begin
      Dec(NewPosTry[1].X);
      Inc(NewPosTry[2].X)
    end
    else
    begin
      Inc(NewPosTry[1].X);
      Dec(NewPosTry[2].X)
    end
  end;
  for I := Low(NewPosTry) to High(NewPosTry) do
    if not CurrentPos.Equal(NewPosTry[I]) then
    begin
      NewPos := TPosition.Create(NewPosTry[I]);
      // Digger
      if L.Map[NewPos.X, NewPos.Y] is TDigger then
        L.Digger.Die;
      // Nothing
      if L.Map[NewPos.X, NewPos.Y] is TNothing then
      begin
        LastDirection := Direction;
        if NewPos.X < CurrentPos.X then
          Direction := Left;
        if NewPos.X > CurrentPos.X then
          Direction := Right;
        if NewPos.Y < CurrentPos.Y then
          Direction := Up;
        if NewPos.Y > CurrentPos.Y then
          Direction := Down;
        case lastDirection of
          Left:
          begin
            if Direction = Down then
              LastTurn := Left;
            if Direction = Up then
              LastTurn := Right
          end;
          Right:
          begin
            if Direction = Down then
              LastTurn := Right;
            if Direction = Up then
              LastTurn := Left
          end;
          Up:
          begin
            if Direction = Left then
              LastTurn := Left;
            if Direction = Right then
             LastTurn := Right
          end;
          Down:
          begin
            if Direction = Left then
              LastTurn := Right;
            if Direction = Right then
              LastTurn := Left
          end;
        end;
        L.SetSprite(NewPos.X, NewPos.Y, Self);
        L.SetSprite(CurrentPos.X, CurrentPos.Y, TNothing.Create);
        Break
      end
    end
end;

{ TGhost90R }

procedure TGhost90R.Move(L: TLevel);
var
  CurrentPos, NewPos: TPosition;
  NewPosTry: array[0..3] of TPosition;
  I: Integer;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  for I := Low(NewPosTry) to High(NewPosTry) do
    NewPosTry[I] := TPosition.Create(CurrentPos);
  if Direction = Left then
  begin
    Dec(NewPosTry[0].X);
    Dec(NewPosTry[1].Y);
    Inc(NewPosTry[2].Y);
    Inc(NewPosTry[3].X)
  end
  else if Direction = Right then
  begin
    Inc(NewPosTry[0].X);
    Inc(NewPosTry[1].Y);
    Dec(NewPosTry[2].Y);
    Dec(NewPosTry[3].X)
  end
  else if Direction = Up then
  begin
    Dec(NewPosTry[0].Y);
    Inc(NewPosTry[1].X);
    Dec(NewPosTry[2].X);
    Inc(NewPosTry[3].Y)
  end
  else if Direction = Down then
  begin
    Inc(NewPosTry[0].Y);
    Dec(NewPosTry[1].X);
    Inc(NewPosTry[2].X);
    Dec(NewPosTry[3].Y)
  end;
  for I := Low(NewPosTry) to High(NewPosTry) do
    if not CurrentPos.Equal(NewPosTry[I]) then
    begin
      NewPos := TPosition.Create(NewPosTry[I]);
      // Digger
      if L.Map[NewPos.X, NewPos.Y] is TDigger then
        L.Digger.Die;
      // Nothing
      if L.Map[NewPos.X, NewPos.Y] is TNothing then
      begin
        if NewPos.X < CurrentPos.X then
          Direction := Left
        else if NewPos.X > CurrentPos.X then
          Direction := Right;
        if NewPos.Y < CurrentPos.Y then
          Direction := Up
        else if NewPos.Y > CurrentPos.Y then
          Direction := Down;
        L.SetSprite(NewPos.X, NewPos.Y, Self);
        L.SetSprite(CurrentPos.X, CurrentPos.Y, TNothing.Create);
        Break
      end
    end
end;

{ TWall }

function TWall.GetBitmap: Integer;
begin
  Result := SpriteWall
end;

{ TDigger }

procedure TDigger.Animate(L: TLevel; X, Y: Integer);
begin
  if Dead then
    Exit;
  Inc(Step);
  case Direction of
    None:
      if Step >= NumDiggerCenterBmps then
        Step := 0;
    Left, Right:
      if Step >= NumDiggerLeftBmps then
        Step := 0;
    Up, Down:
      if Step >= NumDiggerUpBmps then
        Step := 0
  end;
  Invalidate := True
end;

constructor TDigger.Create;
var
  I: Integer;
begin
  inherited;
  //Initializa Boolean array with valid Boolean values
  for I := Low(Keys) to High(Keys) do
    Keys[I] := False
end;

procedure TDigger.Die;
begin
  Dead := True;
  Invalidate := True
end;

function TDigger.GetBitmap: Integer;
const
  LeftBmps:   array[0..NumDiggerLeftBmps   - 1] of Integer = ( 16, 17, 18, 19, 18, 17 );
  RightBmps:  array[0..NumDiggerRightBmps  - 1] of Integer = ( 20, 21, 22, 23, 22, 21 );
  UpBmps:     array[0..NumDiggerUpBmps     - 1] of Integer = ( 24, 25 );
  DownBmps:   array[0..NumDiggerDownBmps   - 1] of Integer = ( 26, 27 );
  CenterBmps: array[0..NumDiggerCenterBmps - 1] of Integer = ( 15, 15, 15, 15, 15, 15, 15, 15, 28, 28, 15, 15, 28, 28, 15, 15, 15, 15, 15, 15, 29, 29, 30, 30, 29, 29, 15, 15, 15, 15 );
begin
  if Dead then
    Result := SpriteDiggerDead
  else
  if (Direction = Left) and (Step < NumDiggerLeftBmps) then
    Result := LeftBmps[step]
  else
  if (Direction = Right) and (Step < NumDiggerRightBmps) then
    Result := RightBmps[step]
  else
  if (Direction = Up) and (Step < NumDiggerUpBmps) then
    Result := UpBmps[step]
  else
  if (Direction = Down) and (Step < NumDiggerDownBmps) then
    Result := DownBmps[step]
  else
    Result := CenterBmps[Step]
end;

procedure TDigger.Move(L: TLevel);
var
  CurrentPos, NewPos: TPosition;
begin
  if Dead then
    Exit;
  CurrentPos := L.GetSpritePosition(Self);
  NewPos := TPosition.Create(CurrentPos);

  Direction := None;
  if Keys[LeftKey] then
  begin
    Dec(NewPos.X);
    Direction := Left
  end
  else
  begin
//    Stone[0] := False;
    if Keys[RightKey] then
    begin
      Inc(NewPos.X);
      Direction := Right
    end
    else
    begin
  //    Stone[1] := False;
      if Keys[UpKey] then
      begin
        Dec(NewPos.Y);
        Direction := Up
      end
      else
      begin
        if Keys[DownKey] then
        begin
          Inc(NewPos.Y);
          Direction := Down
        end
      end
    end
  end;

  //if we're moving to...
  if not CurrentPos.Equal(NewPos) then
  begin
    // ...where a stone is
    if L.Map[NewPos.X, NewPos.Y] is TStone then
    begin
      //if there's nothing to the right of it
      if (NewPos.X > CurrentPos.X) and
         (L.Map[NewPos.X + 1, NewPos.Y] is TNothing) then
      begin
        //if there's a stone to the right of us, move it along
        //if Stone[1] then
        begin
          L.SetSprite(CurrentPos.X + 2, CurrentPos.Y, L.Map[CurrentPos.X + 1, CurrentPos.Y]);
          L.SetSprite(CurrentPos.X + 1, CurrentPos.Y, TNothing.Create)
        end;
        //Stone[1] := not Stone[1];
      end;

      //if there's nothing to the left of it
      if (NewPos.X < CurrentPos.X) and
         (L.Map[NewPos.X - 1, NewPos.Y] is TNothing) then
      begin
        //if there's a stone to the left of us, move it along
        //if Stone[0] then
        begin
          L.SetSprite(CurrentPos.X - 2, CurrentPos.Y, L.Map[CurrentPos.X - 1, CurrentPos.Y]);
          L.SetSprite(CurrentPos.X - 1, CurrentPos.Y, TNothing.Create)
        end;
        //Stone[0] := not Stone[0];
      end
    end;

    // ...where ground, diamond or nothing is
    if (L.Map[NewPos.X, NewPos.Y] is TNothing) or
       (L.Map[NewPos.X, NewPos.Y] is TGround) or
       (L.Map[NewPos.X, NewPos.Y] is TDiamond) then
    begin
      //if a diamond, collect it
      if L.Map[NewPos.X, NewPos.Y] is TDiamond then
      begin
        Inc(L.Collected);
        Inc(L.Score, DiamondPoints)
      end;
      L.SetSprite(NewPos.X, NewPos.Y, Self);
      L.SetSprite(CurrentPos.X, CurrentPos.Y, TBuffer.Create)
    end;

    // ... where the exit is
    if L.Map[NewPos.X, NewPos.Y] is TExit then
      L.ExitLevel;

    // ... where a ghost is
    if L.Map[NewPos.X, NewPos.Y] is TGhost then
      Die
  end;

  Animate(L, NewPos.X, NewPos.Y)
end;

{ TExit }

function TExit.GetBitmap: Integer;
begin
  Result := SpriteExit
end;

{ TChanger }

function TChanger.GetBitmap: Integer;
begin
  Result := SpriteChanger
end;

{ TGravitation }

procedure TGravitation.Move(L: TLevel; X, Y: Integer);
var
  NewX, NewY: Integer;
begin
  NewX := X;
  NewY := Y;

  // Follow gravitation
  if L.Map[X, Y + 1] is TMarker then
    NewY := Y + 1
  else
  begin
    // Falls to the left or to the right
    if (L.Map[X, Y + 1] is TStone) or
       (L.Map[X, Y + 1] is TDiamond) or
       (L.Map[X, Y + 1] is TNothing) then
    begin
      if (L.Map[X - 1, Y + 1] is TMarker) and
         ((L.Map[X - 1, Y] is TNothing) or
          (L.Map[X - 1, Y] is TMarker)) then
      begin
        NewX := X - 1;
        NewY := Y + 1
      end
      else
      begin
        if (L.Map[X + 1, Y + 1] is TMarker) and
           ((L.Map[X + 1, Y] is TNothing) or
            (L.Map[X + 1, Y] is TMarker)) then
        begin
          NewX := X + 1;
          NewY := Y + 1
        end
      end
    end;

    // Changer
    if (L.Map[X, Y + 1] is TChanger) and
       (L.Map[X, Y] is TStone) and
       (L.Map[X, Y + 2] is TMarker) then
      NewY := Y + 2;
  end;

  // Move
  if (NewX <> X) or (NewY <> Y) then
  begin
    if NewY - Y = 2 then
      // Changer
      L.SetSprite(NewX, NewY, TDiamond.Create)
    else
    begin
      L.SetSprite(NewX, NewY, L.Map[X, Y]);
      if L.Map[NewX, NewY] is TUvStone then
        L.SetSprite(NewX, NewY, TStone.Create);
    end;

    L.SetSprite(X, Y, TNothing.Create);

    // Kill digger if necessary
    if L.Map[NewX, NewY + 1] is TDigger then
      L.Digger.Die;

    if L.Map[NewX, NewY + 1] is TGhost then
      TGhost(L.Map[NewX, NewY + 1]).Blast(L)
  end
end;

procedure TGravitation.Simulate(L: TLevel; X, Y: Integer);
var
  NewX, NewY: Integer;
begin
  NewX := X;
  NewY := Y;

  // Nothing
  if L.Map[X, Y + 1] is TNothing then
    NewY := Y + 1
  else
  begin
    // Stone or diamond
    if (L.Map[X, Y + 1] is TStone) or
       (L.Map[X, Y + 1] is TDiamond) then
    begin
      if (L.Map[X - 1, Y + 1] is TNothing) and
         (L.Map[X - 1, Y] is TNothing) then
      begin
        NewX := X - 1;
        NewY := Y + 1
      end
      else
      begin
        if (L.Map[X + 1, Y + 1] is TNothing) and
           (L.Map[X + 1, Y] is TNothing) then
        begin
          NewX := X + 1;
          NewY := Y + 1
        end
      end
    end;

    // Changer
    if (L.Map[X, Y + 1] is TChanger) and
       (L.Map[X, Y] is TStone) and
       (L.Map[X, Y + 2] is TNothing) then
      NewY := Y + 2;
  end;

  // Mark
  if (NewX <> X) or (NewY <> Y) then
    L.SetSprite(NewX, NewY, TMarker.Create);
end;

{ TStone }

function TStone.GetBitmap: Integer;
begin
  Result := SpriteStone
end;

{ TDiamond }

procedure TDiamond.Blink;
begin
  Counter := Succ(Counter) mod NumDiamondBlinks;
  Invalidate := True
end;

function TDiamond.GetBitmap: Integer;
begin
  //original code:
  //Result := LastDiamondBlink - ((Counter + 4) mod NumDiamondBlinks)
  Result := LastDiamondBlink - (Counter mod NumDiamondBlinks)
end;

{ TLevel }

constructor TLevel.Create;
var
  Stream: TStream;
begin
  inherited;
{$ifdef LINK_RESOURCES}
  Stream := GetType.Assembly.GetManifestResourceStream('Level.bin');
{$else}
  Stream := TFileStream.Create('Level.bin', FileMode.Open, FileAccess.Read, FileShare.Read);
{$endif}
  SetLength(Data, Stream.Length);
  Stream.Read(Data, 0, Stream.Length);
end;

procedure TLevel.CreateSprite(X, Y, ID: Integer);
{$ifdef USE_CLASS_REFS}
type
  TSpriteClass = class of TSprite;
var
  ClsType: TSpriteClass;
{$else}
var
  ClsType: System.Type;
{$endif}
begin
  case ID of
    0:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TNothing);
    1:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TStone);
    2:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TGround);
    3:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TGhost180);
    5:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TDiamond);
    6:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TWall);
    7:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TGhost90L);
    9:  ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TUvStone);
    10: ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TDigger);
    11: ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TGhost90LR);
    12: ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TExit);
    14: ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TChanger);
    15: ClsType := {$ifndef USE_CLASS_REFS}TypeOf{$endif}(TGhost90R);
  else
    ClsType := nil
  end;
  if Assigned(ClsType) then
{$ifdef USE_CLASS_REFS}
    SetSprite(X, Y, ClsType.Create);
{$else}
    SetSprite(X, Y, TSprite(Activator.CreateInstance(ClsType)));
{$endif}
end;

procedure TLevel.ExitLevel;
begin
  if (Collected >= Diamonds) and (Number < Pred(HighestLevel)) then
  begin
    Inc(Number);
    Load;
  end
end;

function TLevel.GetSpritePosition(S: TSprite): TPosition;
var
  X, Y: Integer;
begin
  for X := 0 to MapSizeX - 1 do
    for Y := 0 to MapSizeY - 1 do
      if Map[X, Y] = S then
      begin
        Result := TPosition.Create(X, Y);
        Exit;
      end;
  Result := TPosition.Create;
end;

procedure TLevel.Load;
var
  Level: array[0..LevelDataSize-1] of Byte;
  X, Y, I, GhostInfo: Integer;
  Ghost: TGhost;
const
  NewPos: array[0..3] of TDirections = (Down, Up, Right, Left);
begin
  for I := Low(Level) to High(Level) do
    Level[I] := Data[Number * Length(Level) + I];

  for Y := 0 to MapSizeY - 1 do
    for X := 0 to MapSizeX div 2 - 1 do
    begin
      //Create sprite from high nybble info
      CreateSprite(X * 2, Y , Level[Y * MapSizeX div 2 + X] shr 4);
      //Create sprite from low nybble info
      CreateSprite(X * 2 + 1, Y , Level[Y * MapSizeX div 2 + X] and $F);
    end;
{$ifdef USE_CLASS_REFS}
  ShowMessage('TDigger sprite object is actually of type ' + TypeOf(Map[Level[145]][Level[146]-2]).FullName);
{$endif}

  //Digger location is stored in level info
  Digger := Map[Level[145]][Level[146]-2] as TDigger;

  //Ghost locations aren't, so locate them manually
  for I := Low(Ghosts) to High(Ghosts) do
    Ghosts[I] := nil;
  I := 0;
  for Y := 0 to MapSizeY - 1 do
    for X := 0 to MapSizeX - 1 do
    begin
      if Map[X, Y] is TGhost then
      begin
        Ghost := TGhost(Map[X, Y]);
        Ghosts[I] := Ghost;
        //Get ghost info from alternate nybbles of level info
        //(2 ghost's info per byte)
        GhostInfo := Level[148 + (I shr 1)];
        if Odd(I) then
        begin
          GhostInfo := GhostInfo and $F;
          if Ghost is TGhost90LR then
            TGhost90LR(Ghost).LastTurn := Right;
        end
        else //otherwise use high nybble
        begin
          GhostInfo := GhostInfo shr 4;
          if Ghost is TGhost90LR then
            TGhost90LR(Ghost).LastTurn := Left
        end;
        Ghost.Direction := NewPos[GhostInfo];
        Inc(I)
      end
    end;

  Diamonds := Level[147] div $10 * 10 + Level[147] mod $10;
  Collected := 0;
  Time := StartTime
end;

procedure TLevel.SetSprite(X, Y: Integer; S: TSprite);
begin
  Map[X, Y] := S;
  S.Invalidate := True;
end;

procedure TLevel.AnimateDiamonds;
var
  X, Y: Integer;
begin
  for X := 0 to MapSizeX - 1 do
    for Y := 0 to MapSizeY - 1 do
      if Map[X, Y] is TDiamond then
        TDiamond(Map[X, Y]).Blink
end;

procedure TLevel.Tick;
var
  X, Y, I: Integer;
begin
  // Turn buffers into nothing
  for Y := MapSizeY - 1 downto 0 do
    for X := MapSizeX - 1 downto 0 do
      if Map[X, Y] is TBuffer then
        SetSprite(X, Y, TNothing.Create);

  // Digger
  Digger.Move(Self);

  // Stones (normal and UV) and Diamonds
  for Y := MapSizeY - 1 downto 0 do
    for X := MapSizeX - 1 downto 0 do
      if (Map[X, Y] is TStone) or
         (Map[X, Y] is TDiamond) or
         (Map[X, Y] is TUvStone) then
        TGravitation(Map[X, Y]).Simulate(Self, X, Y);

  for Y := MapSizeY - 1 downto 0 do
    for X := MapSizeX - 1 downto 0 do
      if (Map[X, Y] is TStone) or
         (Map[X, Y] is TDiamond) or
         (Map[X, Y] is TUvStone) then
        TGravitation(Map[X, Y]).Move(Self, X, Y);

  //Ghosts
  for I := Low(Ghosts) to High(Ghosts) do
    if Assigned(Ghosts[I]) then
      Ghosts[I].Move(Self);

  if Time > 0 then
    Dec(Time);
  if Time = 0 then
    Digger.Die;
end;

end.
