unit DiggerMainForm;

{$ifdef USE_RANGE_CHECKING}
  {$RANGECHECKS ON}
{$else}
  {$RANGECHECKS OFF}
{$endif}

//Resources cannot be linked in yet, so we'll leave them as separate files
{$ifdef LINK_RESOURCES}
  //Change these directives once managed resource linking is properly supported
  //{$R Digger.ico}
  //{$R Sprite.png}
  //{$R Font.png}
{$endif}

interface

uses
  DiggerFont,
  DiggerLevel,
  Borland.Delphi.Types,
  Oblong.Types,
  Oblong.Vcl.Controls;

const
  NumSprites = 34;

type
  TMainForm = class(TForm)
  private
    Sprites: array[0..NumSprites-1] of TBitmap;
    Font: TFont;
    Timer: TTimer;
    Tick: Integer; //defaults to 0
    Level: TLevel;
    Lives: Integer;
    KeysRelease: array[0..NumDirKeys-1] of Boolean; //which keys to release on next timer tick
  protected
    procedure OnKeyDown(E: TKeyEventArgs); override;
    procedure OnKeyUp(E: TKeyEventArgs); override;
    procedure OnPaint(E: TPaintEventArgs); override;
    procedure OnPaintBackground(E: TPaintEventArgs); override;
    procedure OnTimer(Sender: TObject; E: TEventArgs);
  public
    constructor Create;
    function MagnifyRectangle(X, Y, DX, DY: Integer): TRect;
    function MagnifyBitmap(Source: TBitmap): TBitmap;
  end;

var
  MainForm: TMainForm;

implementation

{ TMainForm }

const
  Magnification  = 2;
  NumLives       = 20;
  TimerInterval  = 30 {msec};
  ClientWidth    = 320;
  ClientHeight   = 256;
  HeightOffset   = 32;
  CellWidth      = 16;
  CellHeight     = 16;
  FontUnitWidth  = 8;
  FontUnitHeight = 8;

constructor TMainForm.Create;
var
  AllSprites: TBitmap;
  Cursor: TRect;
  I: Integer;
begin
  inherited;
{$ifdef LINK_RESOURCES}
  Icon := TIcon.Create(GetType.Assembly.GetManifestResourceStream('Digger.ico'));
{$else}
  Icon := TIcon.Create('Digger.ico');
{$endif}
  Text := 'Lutz Roeder''s .NET Digger, translated to Delphi for .NET by Brian Long';
  FormBorderStyle := TFormBorderStyle.FixedDialog;
  StartPosition := TFormStartPosition.CenterScreen;
  MaximizeBox := False;
  Lives := NumLives;
  ClientSize := MagnifyRectangle(0, 0, ClientWidth, ClientHeight).Size;
{$ifdef LINK_RESOURCES}
  Font := TFont.Create(MagnifyBitmap(TBitmap.Create(GetType.Assembly.GetManifestResourceStream('Font.png'))), MagnifyRectangle(0, 0, 8, 8).Size);
  AllSprites := MagnifyBitmap(TBitmap.Create(GetType.Assembly.GetManifestResourceStream('Sprite.png')));
{$else}
  Font := TFont.Create(MagnifyBitmap(TBitmap.Create('Font.png')), MagnifyRectangle(0, 0, FontUnitWidth, FontUnitHeight).Size);
  AllSprites := MagnifyBitmap(TBitmap.Create('Sprite.png'));
{$endif}
  Cursor := MagnifyRectangle(0, 0, CellWidth, CellHeight);
  for I := 0 to NumSprites - 1 do
  begin
    Sprites[I] := AllSprites.Clone(Cursor, TPixelFormat.Format32bppArgb);
    Cursor.X := Cursor.X + Cursor.Width;
  end;

  Level := TLevel.Create;
  Level.Load;

  Timer := TTimer.Create;
  Timer.Interval := TimerInterval;
  Timer.Add_Tick(OnTimer);
  Timer.Start;
end;

function TMainForm.MagnifyBitmap(Source: TBitmap): TBitmap;
var
  X, Y, IX, IY: Integer;
begin
  Result := TBitmap.Create(Source.Width * Magnification, Source.Height * Magnification);
  for X := 0 to Source.Width - 1 do
    for Y := 0 to Source.Height - 1 do
      for IX := 0 to Magnification - 1 do
        for IY := 0 to Magnification - 1 do
          Result.SetPixel(X * Magnification + IX, Y * Magnification + IY, Source.GetPixel(X, Y));
end;

function TMainForm.MagnifyRectangle(X, Y, DX, DY: Integer): TRect;
begin
  Result := TRect.Create(X * Magnification, Y * Magnification, DX * Magnification, DY * Magnification);
end;

procedure TMainForm.OnKeyDown(E: TKeyEventArgs);
begin
  if E.KeyCode = TKeys.Left then
    Level.Digger.Keys[LeftKey] := True;
  if E.KeyCode = TKeys.Right then
    Level.Digger.Keys[RightKey] := True;
  if E.KeyCode = TKeys.Up then
    Level.Digger.Keys[UpKey] := True;
  if E.KeyCode = TKeys.Down then
    Level.Digger.Keys[DownKey] := True;
  if (E.KeyCode = TKeys.Escape) and (Lives > 0) then
  begin
    Dec(Lives);
    Level.Load
  end;
  //Original version allowed you to proceed to next level (by pressing Home)
  //even if you were dead and had no lives left
  if (E.KeyCode = TKeys.Home) and (Level.Number < HighestLevel - 1) and (Lives > 0) then
  begin
    Inc(Level.Number);
    Level.Load
  end
end;

procedure TMainForm.OnKeyUp(E: TKeyEventArgs);
begin
  if E.KeyCode = TKeys.Left then
  begin
    Level.Digger.Keys[LeftKey] := False;
    KeysRelease[LeftKey] := True
  end;
  if E.KeyCode = TKeys.Right then
  begin
    Level.Digger.Keys[RightKey] := False;
    KeysRelease[RightKey] := True
  end;
  if E.KeyCode = TKeys.Up then
  begin
    Level.Digger.Keys[UpKey] := False;
    KeysRelease[UpKey] := True
  end;
  if E.KeyCode = TKeys.Down then
  begin
    Level.Digger.Keys[DownKey] := False;
    KeysRelease[DownKey] := True
  end;
end;

procedure TMainForm.OnPaint(E: TPaintEventArgs);
var
  R, S: TRect;
  Canvas: TBitmap;
  Graphics: TGraphics;
  Blue, Cyan, Red: TSolidBrush;
  X, Y: Integer;
begin
  R := E.ClipRectangle;
  if (R.Width = 0) or (R.Height = 0) then
    Exit;
  Canvas := TBitmap.Create(R.Width, R.Height, E.Graphics);
  Graphics := TGraphics.FromImage(Canvas);
  Graphics.TranslateTransform(-R.X, -R.Y);
  Graphics.SetClip(R);

  Blue := TSolidBrush.Create(TColor.FromArgb($04, $02, $8F));
  Graphics.FillRectangle(Blue, MagnifyRectangle(0, 0, ClientWidth, 8));
  Graphics.FillRectangle(Blue, MagnifyRectangle(0, 24, ClientWidth, 8));

  Cyan := TSolidBrush.Create(TColor.FromArgb($2D, $E7, $C0));
  Graphics.FillRectangle(Cyan, MagnifyRectangle(0, 2, ClientWidth, 4));
  Graphics.FillRectangle(Cyan, MagnifyRectangle(0, 26, ClientWidth, 4));

  Red := TSolidBrush.Create(TColor.FromArgb($92, $02, $05));
  Graphics.FillRectangle(Red, MagnifyRectangle(0, 8, ClientWidth, 16));

  Font.Reset;
  Font.Newline;
{$ifdef USE_OBJECT_BOXING}
  Font.Write(Graphics, '  CAVE:  ' + Int32(Level.Number + 1).ToString('D2') +
    ' TIME:  ' + Int32(Level.Time).ToString('D5') +
    ' DIAMONDS:  ' + Int32(Level.Diamonds).ToString('D2'));
  Font.Newline;
  Font.Write(Graphics, '  LIVES: ' + Int32(Lives).ToString('D2') +
    ' SCORE: ' + Int32(Level.Score).ToString('D5') +
    ' COLLECTED: ' + Int32(Level.Collected).ToString('D2'));
{$else}
  Font.Write(Graphics, Format('  CAVE:  %.2d TIME:  %.5d DIAMONDS:  %.2d',
    [Level.Number + 1, Level.Time, Level.Diamonds]));
  Font.Newline;
  Font.Write(Graphics, Format('  LIVES: %.2d SCORE: %.5d COLLECTED: %.2d',
    [Lives, Level.Score, Level.Collected]));
{$endif}
  Font.Newline;

  //Draw cells within the clipping rectangle
  for Y := 0 to MapSizeY - 1 do
    for X := 0 to MapSizeX - 1 do
    begin
      S := MagnifyRectangle(X * CellWidth, Y * CellHeight + HeightOffset, CellWidth, CellHeight);
      if R.IntersectsWith(S) then
        Graphics.DrawImageUnscaled(Sprites[Level.Map[X, Y].GetBitmap], S.X, S.Y);
    end;

  Graphics.Dispose;
  E.Graphics.DrawImageUnscaled(Canvas, R.X, R.Y);
  Canvas.Dispose;
end;

procedure TMainForm.OnPaintBackground(E: TPaintEventArgs);
begin
  //Do nothing since we are drawing the whole form client area
end;

procedure TMainForm.OnTimer(Sender: TObject; E: TEventArgs);
var
  X, Y: Integer;
  Sprite: TSprite;
  R: TRect;
begin
  //Animate diamonds twice as often as we update the digger, stones, etc.
  Inc(Tick);
  Level.AnimateDiamonds;
  if not Odd(Tick) then
  begin
    if KeysRelease[LeftKey] then
    begin
      Level.Digger.Keys[LeftKey] := False;
      KeysRelease[LeftKey] := False
    end;
    if KeysRelease[RightKey] then
    begin
      Level.Digger.Keys[RightKey] := False;
      KeysRelease[RightKey] := False
    end;
    if KeysRelease[UpKey] then
    begin
      Level.Digger.Keys[UpKey] := False;
      KeysRelease[UpKey] := False
    end;
    if KeysRelease[DownKey] then
    begin
      Level.Digger.Keys[DownKey] := False;
      KeysRelease[DownKey] := False
    end;
    Level.Tick;
  end;
  //Update screen each timer tick
  Invalidate(MagnifyRectangle(0, 8, ClientWidth, CellHeight));
  for Y := 0 to MapSizeY - 1 do
    for X := 0 to MapSizeX - 1 do
    begin
      Sprite := Level.Map[X, Y];
      if Sprite.Invalidate then
      begin
        R := MagnifyRectangle(X * CellWidth, Y * CellHeight + HeightOffset, CellWidth, CellHeight);
        Invalidate(R);
        Sprite.Invalidate := False
      end
    end;
  Update;
end;

end.
